@@ -287,18 +287,25 @@ class Client(suds.client.Client):
     def getEntity(self, obj):
         """Get the corresponding :class:`icat.entity.Entity` for an object.
 
-        if obj is a `fieldSet`, return the list of fields.  If obj is
-        any other Suds instance object, create a new entity object
+        if obj is a `fieldSet`, return a tuple of the fields.  If obj
+        is any other Suds instance object, create a new entity object
         with :meth:`~icat.client.Client.new`.  Otherwise do nothing
         and return obj unchanged.
         
         :param obj: either a Suds instance object or anything.
         :type obj: :class:`suds.sudsobject.Object` or any type
         :return: the new entity object or obj.
-        :rtype: :class:`list` or :class:`icat.entity.Entity` or any type
+        :rtype: :class:`tuple` or :class:`icat.entity.Entity` or any type
+
+        .. versionchanged:: 0.18.0
+            add support of `fieldSet`.
+
+        .. versionchanged:: 0.18.1
+            changed the return type from :class:`list` to
+            :class:`tuple` in the case of `fieldSet`.
         """
         if obj.__class__.__name__ == 'fieldSet':
-            return obj.fields
+            return tuple(obj.fields)
         elif isinstance(obj, suds.sudsobject.Object):
             return self.new(obj)
         else:
